unit TPUnit;     { Arm Version 1.0 with diagram, May 26, 1989 }
                 { revised with new graphics routines, Feb. 28, 1990 }
interface

const left_ = 203; right = 205; up = 200; down = 208; PgUp = 201; PgDn = 209;
      Ins = 210; Del = 211; EndKey = 207; Home = 199;
      CLeft =  243; Cright = 244; CEnd = 245;
      f1 = 187; f2 = 188; f3 = 189; f4 = 190; f5 = 191;
      f6 = 192; f7 = 193; f8 = 194; f9 = 195; f10 = 196;
      Altf1 = 232;Altf2 = 233;Altf3 = 234;Altf4 = 235;Altf5 = 236;
      Altf6 = 237;Altf7 = 238;Altf8 = 239;Altf9 = 240;Altf10 = 241;
      f11 = 261; F12 = 262;
      EscKey = $1b;

var g1a,g1b,g1c,g1d,g1e,s1a,s1b,s1c,s1d,s1e,   { gain constants }
    g2a,g2b,g2c: integer;
    theta,phi: integer;
    delta: integer;
    halfside: integer;       { half of one side of box }
    TAx,            { x position of target }
    TAy,            { y position of target }
    TAz: integer;    { z position of target }
    stage: integer;
    editing: boolean;
    p1a,p1b,p1c,p1d,p1e,e1a,e1b,
    e1c,e1d,e1e,o1a,o1b,o1c,o1d,
    o1e,r1a,r1b,r1c,r1d,r1e: integer;   { lower order control }
    p2a,p2b,p2c,e2a,e2b,e2c,
    o2a,o2b,o2c,r2a,r2b,r2c: integer;   { higher order control }
    purpose: boolean;
    purx,pury: integer;
    elapsedtime: double;

procedure initialize;
procedure iterate;
procedure saveedit;

implementation

uses armv1un, math;

const scale = 4096;
const one80 = 2048;   {180 degrees}
      maxmin1 = one80 - 1;
const three4pi = (one80 + one80 + one80) div 4;
const rightangle = one80 div 2;

type poly3dtype = array[1..50] of
                 record
                  x3d,y3d,z3d: integer;
                 end;


     poly2dptrtype = ^poly2dtype;

     sincostype = array[-one80..maxmin1] of integer;
     invtantype = array[0..scale] of integer;

var
    rf,             { radius, shoulder to finger }
    FIx,            { x position of finger }
    FIy,            { y position of finger }
    FIz,            { z position of finger }
    REx,            { right eye x coordinate }
    REy,            { right eye y coordinate }
    REz,            { right eye z coordinate }
    LEx,            { left_ eye x coordinate }
    LEy,            { left_ eye y coordinate }
    LEz,            { left_ eye z coordinate }
    RExr,            { right eye rotated x coordinate }
    REyr,            { right eye rotated y coordinate }
    REzr,            { right eye rotated z coordinate }
    LExr,            { left_ eye rotated x coordinate }
    LEyr,            { left_ eye rotated y coordinate }
    LEzr,            { left_ eye rotated z coordinate }
    ELx,            { x coordinate of elbow }
    ELy,            { y coordinate of elbow }
    ELz,            { z coordinate of elbow }
    SHx,            { x coordinate of shoulder }
    SHy,            { y coordinate of shoulder }
    SHz,            { z coordinate of shoulder }
    ra,             { length of upper and lower arms }
    t1,             { angle from nadir to upper arm }
    t2,             { angle from projection of upper arm to lower arm }
    t3,             { angle from shoulder to finger }
    az,             { azimuth angle of arm }
    TLRd,           { tan depth angle to target }
    FLRd,           { tan depth angle to finger }

    TRx,            { azimuth angle, right eye to target }
    TLx,            { azimuth angle, left_  eye to target }
    FRx,            { azimuth angle, right eye to finger }
    FLx,            { azimuth angle, left_  eye to finger }

    TRy_,           { vertical angle, right eye to target. Conflict with "try" }
    TLy,            { vertical angle, left_  eye to target }
    FRy,            { vertical angle, right eye to finger }
    FLy,            { vertical angle, left_  eye to finger }

    TXL,TYL,TZL,FXL,FYL,FZL,   { positions relative to eyes }
    TXR,TYR,TZR,FXR,FYR,FZR: integer;


    pf: integer;       { angle from horizontal, eyes to finger }
    temp,at1,at3,az1: integer;

    i1a,i1b,i1c,i1d,i1e,i2a,i2b,i2c: longint;
    lim1a,lim1b,lim1c,lim1d,lim1e,lim2a,lim2b,lim2c: longint;
    errlim: integer;

    paramfile: file of integer;

    x,y,z,xp,yp,zp: integer;
    ch: char;

    cube3,body3,head3,head3r,eyes3,eyes3r,arm3,target3: poly3dtype;
    cube2,body2,head2,eyes2,arm2,target2: poly2dtype;
    gazex,gazey,phi2,theta2,d: integer;
    headwidth,headheight,neck,eyesize: integer;

    sine,cosine: sincostype;
    invtan: invtantype;
    invtanfile: file of invtantype;
    sincosfile: file of sincostype;

    toptrace,middletrace,bottomtrace:integer;
    tottime: integer;
    tempt: integer;
    pr,pt: integer;
    usediagram: boolean;
    vsize,hsize,vcenter,hcenter: integer;

procedure dots(n: integer);
var i,x,y: integer;
begin
 if n = 0 then exit;
 x := wherex; y := wherey;
 for i := 1 to n do writestr('.');
 gotoxy(x,y);
end;

procedure blanks(n: integer);
var i,x,y: integer;
begin
 if n = 0 then exit;
 x := wherex; y := wherey;
 for i := 1 to n do writestr(' ');
 gotoxy(x,y);
end;

function getcod: byte;
begin
 getcod := readkey;
end;

procedure initscreen;
begin
 hsize := getmaxx; vsize := getmaxy;
 vcenter := vsize div 2; hcenter := hsize div 2;
 setviewport(0,361,hsize,vsize);
 clearviewport;
end;


procedure makesincos;
var i: integer;
    u,v: real;
begin
 clrscr; writelnstr(' '); writelnstr(
'     Trigonometric tables not found: creating SINCOS and INVTAN');
 for i := -one80 to one80 - 1 do
 begin
  u := scale * cos(i * pi/one80);
  v := scale * sin(i * pi/one80);
  cosine[i] := round(u);
  sine[i] := round(v);
 end;
 for i := 0 to scale do
  invtan[i] := round(one80/pi*arctan(1.0*i/scale));
end;

procedure getfunctions;
begin
 assign(sincosfile,'sincos');
 {$i-}
 reset(sincosfile);
 {$i+}
 if IOResult <> 0 then
  begin
   makesincos;
   rewrite(sincosfile);
   write(sincosfile,sine,cosine);
   close(sincosfile);
  end
 else
  begin
   read(sincosfile,sine,cosine);
   close(sincosfile);
  end;
 assignfile(invtanfile,'invtan');
 {$i-}
 reset(invtanfile);
 {$i+}
 if IOResult <> 0 then
  begin
   rewrite(invtanfile);
   write(invtanfile,invtan);
   close(invtanfile);
  end
 else
  begin
   read(invtanfile,invtan);
   close(invtanfile);
  end;
end;

function dmd(m1,m2,d: integer): integer;
var temp:longint;
begin
 temp := m1*m2;
 if d <> 0 then
 dmd := temp div d
 else
 dmd := maxint;
end;

function findangle(x,y: integer): integer;
begin
 findangle := round(2048/pi*arctan2(y,x));
end;

procedure rotatelook(var poly1: poly3dtype;   { rotate from head3/eyes3 to }
                     var  poly2: poly3dtype;  { head3r/eyes3r }
                     n,theta2,phi2: integer);
var i: integer;
    ca,sa: integer;
begin
 for i := 1 to n do
  begin
   with poly1[i] do
   begin
    x := x3d;
    y := y3d - REx-REx;
    z := z3d + halfside;
   end;
   ca := cosine[theta2]; sa := sine[theta2];
   yp := dmd(ca,y,scale) + dmd(sa,z,scale);
   zp := -dmd(sa,y,scale) + dmd(ca,z,scale);
   y := yp; z := zp;
   ca := cosine[phi2]; sa := sine[phi2];
   xp := dmd(ca,x,scale) + dmd(sa,z,scale);
   zp := -dmd(sa,x,scale) + dmd(ca,z,scale);
   with poly2[i] do
   begin
    x3d := xp;
    y3d := REx + REx + yp;
    z3d := -halfside + zp;
   end;
  end;
end;

procedure projectxy(var   poly1: poly3dtype;
                    var  poly2d: poly2dtype;
                    n,theta,phi,d: integer);
var i,ca,sa: integer;
    p1: poly3dtype;
    x0,y0: integer;
begin
 if theta < -one80 then theta := theta + one80 + one80;
 if theta >= one80 then theta := theta - one80 - one80;
 x0 := (hcenter + hcenter + hcenter) div 2;
 y0 := round(0.6*vcenter); // - vsize shr 5;
 for i := 1 to n do
  begin
   with poly1[i] do
    begin
     x := x3d;
     y := y3d;
     z := z3d;
    end;

   ca := cosine[phi]; sa := sine[phi];
   xp := dmd(ca,x,scale) + dmd(sa,z,scale);
   zp := -dmd(sa,x,scale) + dmd(ca,z,scale);

   x := xp; z := zp;

   ca := cosine[theta]; sa := sine[theta];
   yp := dmd(ca,y,scale) + dmd(sa,z,scale);
   zp := -dmd(sa,y,scale) + dmd(ca,z,scale);

   with p1[i] do
   begin
    x3d := xp;
    y3d := yp;
    z3d := zp + d;
   end;
   with poly2d[i],p1[i] do
   if(z3d <> -500) then
    begin
     x2d := x0 + dmd(x3d,500,(500 + z3d));
     y2d := y0 - dmd(y3d,500,(500 + z3d));
    end;
  end;
end;

procedure rotateimage(var Ax,Ay: integer;      { visual angles, results}
                          Vx,Vy,Vz: integer;   { coordinates, finger or target}
                          th2,ph2: integer);   { theta = x axis, phi = y axis}
var sa,ca,y,z,xp,yp,zp: integer;
begin
 x := Vx;
 y := Vy;
 z := Vz;

 ca := cosine[-th2]; sa := sine[-th2];

 yp :=  dmd(y,ca,scale) + dmd(z,sa,scale);
 zp := -dmd(y,sa,scale) + dmd(z,ca,scale);

 y := yp; z := zp;

 ca := cosine[-ph2]; sa := sine[-ph2];

 xp:=  dmd(x,ca,scale) + dmd(z,sa,scale);
 zp := -dmd(x,sa,scale) + dmd(z,ca,scale);
 x := xp; z := zp;

 Ax := findangle(z,x);
 Ay := findangle(z,y);
end;



procedure makecube;
var poly: poly3dtype;
begin
 with poly[01] do begin x3d := -halfside; y3d := -halfside;z3d := -halfside end;
 with poly[02] do begin x3d := -halfside; y3d :=  halfside;z3d := -halfside end;
 with poly[03] do begin x3d :=  halfside; y3d :=  halfside;z3d := -halfside end;
 with poly[04] do begin x3d :=  halfside; y3d := -halfside;z3d := -halfside end;
 with poly[05] do begin x3d := -halfside; y3d := -halfside;z3d := -halfside end;
 with poly[06] do begin x3d := -halfside; y3d := -halfside;z3d := -halfside end;
 with poly[07] do begin x3d := -halfside; y3d := -halfside;z3d :=  halfside end;
 with poly[08] do begin x3d := -halfside; y3d :=  halfside;z3d :=  halfside end;
 with poly[09] do begin x3d :=  halfside; y3d :=  halfside;z3d :=  halfside end;
 with poly[10] do begin x3d :=  halfside; y3d := -halfside;z3d :=  halfside end;
 with poly[11] do begin x3d :=  halfside; y3d := -halfside;z3d := -halfside end;
 cube3 := poly;
end;

procedure makebody;
var poly: poly3dtype;
    du: integer;
begin
 du := headwidth div 2;
 with poly[01] do begin x3d :=    0; y3d :=    0;z3d := -halfside end;
 with poly[02] do begin x3d :=  -SHx; y3d :=    0;z3d := -halfside end;
 with poly[03] do begin x3d :=    0; y3d := -halfside; z3d := -halfside end;
 with poly[04] do begin x3d :=   SHx; y3d :=    0;z3d := -halfside end;
 with poly[05] do begin x3d :=    0; y3d :=    0;z3d := -halfside end;
 with poly[06] do begin x3d :=    0; y3d :=   0; z3d := -halfside end;
 with poly[07] do begin x3d :=    0; y3d :=  du; z3d := -halfside end;
 body3 := poly;
end;

procedure makehead;
var poly: poly3dtype;
    du,dv: integer;
begin
 du := headwidth div 2;
 dv := headheight;
 with poly[01] do begin x3d :=    0; y3d :=  neck; z3d := -halfside end;
 with poly[02] do begin x3d :=   du; y3d :=  neck; z3d := -halfside end;
 with poly[03] do begin x3d :=   du; y3d :=  neck + dv; z3d := -halfside end;
 with poly[04] do begin x3d :=  -du; y3d :=  neck + dv; z3d := -halfside end;
 with poly[05] do begin x3d :=  -du; y3d :=  neck; z3d := -halfside end;
 with poly[06] do begin x3d :=    0; y3d :=  neck; z3d := -halfside end;
 head3 := poly;
end;

procedure makeeyes;
var poly: poly3dtype;
    e: integer;
begin
 e := eyesize div 2;
 with poly[01] do begin x3d :=  LEx - e; y3d :=  LEy - e; z3d := LEz end;
 with poly[02] do begin x3d :=  LEx - e; y3d :=  LEy + e; z3d := LEz end;
 with poly[03] do begin x3d :=  LEx + e; y3d :=  LEy + e; z3d := LEz end;
 with poly[04] do begin x3d :=  LEx + e; y3d :=  LEy - e; z3d := LEz end;
 with poly[05] do begin x3d :=  LEx - e; y3d :=  LEy - e; z3d := LEz end;
 with poly[06] do begin x3d :=  REx - e; y3d :=  REy - e; z3d := LEz end;
 with poly[07] do begin x3d :=  REx - e; y3d :=  REy + e; z3d := REz end;
 with poly[08] do begin x3d :=  REx + e; y3d :=  REy + e; z3d := REz end;
 with poly[09] do begin x3d :=  REx + e; y3d :=  REy - e; z3d := REz end;
 with poly[10] do begin x3d :=  REx - e; y3d :=  REy - e; z3d := REz end;
 eyes3 := poly;
end;

procedure seteyes;
begin
 with eyes3r[01] do
  begin
   LExr := x3d + 8;
   LEyr := y3d + 8;
   LEzr := z3d;
  end;
 with eyes3r[06] do
  begin
   RExr := x3d + 8;
   REyr := y3d + 8;
   REzr := z3d;
  end;
end;


procedure maketarget(theta,phi,d: integer);
var ts: integer;
begin
  ts := dmd(24,vsize,348);
  with target3[01] do begin x3d :=  TAx; y3d :=  TAy - ts; z3d := TAz end;
  with target3[02] do begin x3d :=  TAx - ts; y3d := TAy + ts ;z3d := TAz end;
  with target3[03] do begin x3d :=  TAx + ts; y3d :=  TAy + ts;z3d := TAz end;
  with target3[04] do begin x3d :=  TAx; y3d :=  TAy - ts; z3d := TAz end;
  projectxy(target3,target2,4,theta,phi,d);
end;

procedure setlimits;
var u: longint;
 begin
  lim1a := longint(50) * maxmin1;
  lim1b := longint(50) * maxmin1;
  lim1c := longint(100) * maxmin1;
  u := longint(maxmin1) div 2;
  lim1d := u * g1d;
  lim1e := u * g1e;
  lim2a := u * g2a;
  lim2b := u * g2b;
  lim2c := longint(maxmin1) * g2c;
  errlim := 500;
 end;

procedure initialize;
begin
 initscreen;
 halfside := vsize div 2; delta := 50;
 headheight := halfside div 2;
 headwidth := halfside div 2;
 eyesize := halfside div 10;
 neck := halfside div 4;
 d := 1000;
 ra :=  dmd(5,halfside,4);
 REx := headwidth div 4;
 REy := dmd(5,halfside,8);
 REz := -halfside + halfside div 8;
 LEx := -REx; LEy := REy; LEz := REz;
 SHx :=  halfside div 2; SHy :=   0; SHz := -halfside;
 FIx :=  TAx; FIy := TAy; FIz := TAz;
 t1 := -rightangle div 2;   t2 := 512; az := 0;
 r1a := 0; r1b := 0; r1c := 0;
 o1a := 0; o1b := 0; o1c := 0; o1d := 100; o1e := 100;
 o2a := 0; o2b := 0; o2c := 0;
 i1a := 0; i1b := 0; i1c := 0; i1d := 0; i1e := 0;
 i2a := 0; i2b := 0; i2c := 0;
 setlimits;
 editing := false;
 loaddiagram;
 tottime := 20;
end;

procedure environment;
begin
 t3 := t1 + t2 div 2; { horiz to shoulder-finger line }

{FIND X,Y,Z COORDINATES OF FINGERTIP }

 rf :=  (dmd(ra + ra,cosine[t2 div 2],scale));  { radius, shoulder-finger }

 FIy := SHy + dmd(rf,sine[t3],scale);      { y coordinate of finger }
 temp := dmd(rf,cosine[t3],scale);
 FIx := SHx + dmd(temp,sine[az],scale);      { x coordinate of finger }
 FIz := SHz + dmd(temp,cosine[az],scale);    { z coordinate of finger }
 if FIz < LEz + 25 then FIz := LEz + 25;
 if FIz < REz + 25 then FIz := REz + 25;

{FIND X,Y,Z COORDINATES OF ELBOW }

 ELy := SHy + dmd(ra,sine[t1],scale);
 temp := dmd(ra,cosine[t1],scale);
 ELx := SHx + dmd(temp,sine[az],scale);
 ELz := SHz + dmd(temp,cosine[az],scale);

  { FILL IN POLYGON FOR ARM }

 arm3[1].x3d := SHx; arm3[1].y3d := SHy; arm3[1].z3d := SHz;
 arm3[2].x3d := ELx; arm3[2].y3d := ELy; arm3[2].z3d := ELz;
 arm3[3].x3d := FIx; arm3[3].y3d := FIy; arm3[3].z3d := FIz;

 { CALCULATE LOCAL VARIABLES FOR COMPUTING EYE ANGLES }

 TXL := TAx - LExr; TXR := TAx - RExr;
 TYL := TAy - LEyr; TYR := TAy - REyr;
 TZL := TAz - LEzr; TZR := TAz - REzr;

 FXL := FIx - LExr; FXR := FIx - RExr;
 FYL := FIy - LEyr; FYR := FIy - REyr;
 FZL := FIz - LEzr; FZR := FIz - REzr;

  { FIND ANGLES, FINGER and TARGET, RIGHT EYE }

 rotateimage(TRx,TRy_,TXR,TYR,TZR,theta2,phi2);
 rotateimage(FRx,FRy,FXR,FYR,FZR,theta2,phi2);

 { FIND ANGLES, FINGER and TARGET, left_ EYE }

 rotateimage(TLx,TLy,TXL,TYL,TZL,theta2,phi2);
 rotateimage(FLx,FLy,FXL,FYL,FZL,theta2,phi2);

 { FIND DEPTH ANGLES, EYES TO TARGET AND FINGER, radians }

 FLRd := FLx - FRx;
 TLRd := TLx - TRx;

 end;

procedure azimuth1;
begin
 if r1a > 768 then r1a := 768
 else
 if r1a < -768 then r1a := -768;
 p1a := az;
 e1a := r1a - p1a;
 i1a:= i1a + (100*g1b*e1a - i1a) div s1a;
 o1a := i1a div 100;
end;

procedure vertical1;
begin
 if r1b < -1024 then r1b := -1024
 else
 if r1b > 1024 then r1b := 1024;
 p1b := t1 + 400;
 e1b := r1b  - p1b;
 i1b := i1b + (100*g1b*e1b - i1a) div s1b;
 o1b := i1b div 100;
end;

procedure distance1;
begin
 if r1c > 768 then r1c := 768
 else
 if r1c < -1024 then r1c := -1024;
 p1c := t2 - 768;   { kinesthetic perception of shoulder-finger distance }
 e1c :=  -r1c - p1c;
 i1c := i1c + (100*g1c*e1c - i1c) div s1c;
 o1c := i1c div 100;
 end;

procedure gaze1x;
begin
 r1d := 0;
 p1d := TRx;         { perceived x target deviation from gaze angle }
 e1d := r1d - p1d;
 i1d := i1d - e1d;
 o1d := dmd(i1d,g1d,30);
end;

procedure gaze1y;
begin
 r1e := 0;
 p1e := TRy_;         { perceived y target deviation from gaze angle }
 e1e := r1e - p1e;
 i1e := i1e - e1e;
 o1e := dmd(i1e,g1e,30);
end;

procedure link1to0;
begin
 t1 := o1b;      { vertical }
 az := o1a;      { azimuth }
 t2 := o1c;      { distance }
 gazex := o1d;   { gaze x }
 gazey := o1e;   { gaze y }
end;

procedure azimuth2;
begin
 if purpose then
  r2a := round(120*cos(elapsedtime))
 else r2a := 0;
 p2a := TRx - FRx;
 e2a := r2a - p2a;
 i2a := i2a + e2a;
 o2a := -dmd(i2a,g2a,100);
end;

procedure vertical2;
begin
 if purpose then
  r2b := round(120*sin(elapsedtime))
 else r2b := 0;
 p2b := FRy - TRy_;
 e2b := r2b - p2b;
 i2b := i2b + e2b;
 o2b := dmd(i2b,g2b,100);
end;

procedure distance2;
begin
 r2c := 0;
 p2c := dmd(REx - LEx,4096,TLRd) - dmd(REx - LEx,4096,FLRd);
 e2c := r2c - p2c;
 i2c := i2c + e2c;
 o2c := -dmd(i2c,g2c,100);
end;

procedure link2to1;
begin
 r1a := o2a;              { finger azimuth}
 r1b := o2b + o2c div 2;  { finger pitch  }
 r1c := o2c;              { fdinger distance }
end;

procedure computeenvironment;
begin
 projectxy(cube3,cube2,11,theta,phi,d);
 projectxy(body3,body2,7,theta,phi,d);
end;

procedure drawenvironment;
begin
 drawpoly(11,cube2); // draw environment
 drawpoly(7,body2);
end;

procedure computearmhead;
begin
 projectxy(cube3,cube2,11,theta,phi,d);  // set environment
 projectxy(body3,body2,7,theta,phi,d);
 projectxy(arm3,arm2,3,theta,phi,d);     // set head, arm, target
 projectxy(head3r,head2,6,theta,phi,d);
 projectxy(eyes3r,eyes2,10,theta,phi,d);
 maketarget(theta,phi,d);
end;

procedure drawarmhead;
begin
 drawpoly(3,arm2); // draw arm, head, eyes, target
 drawpoly(10,eyes2);
 drawpoly(6,head2);
 drawpoly(4,target2);
end;

procedure  filldiagram;
var s: string;
begin
 with armv1fm do
 begin
  str(TAx:04,s);
  outtextxy1(36,14,s);
  str(TAy:04,s);
  outtextxy1(36,28,s);
  str(TAz:04,s);
  outtextxy1(36,42,s);
  str(FIx:4,s);
  outtextxy1(93,14,s);
  str(FIy:4,s);
  outtextxy1(93,28,s);
  str(FIz:4,s);
  outtextxy1(93,42,s);
  str(p2b:4,s);
  outtextxy1(141,93,s);
  str(e2b:4,s);
  outtextxy1(170,93,s);
  str(o2b:4,s);
  outtextxy1(195,93,s);
  str(p2c:4,s);
  outtextxy1(241,93,s);
  str(e2c:4,s);
  outtextxy1(268,93,s);
  str(o2c:4,s);
  outtextxy1(295,93,s);
  str(p2a:4,s);
  outtextxy1(337,93,s);
  str(e2a:4,s);
  outtextxy1(365,93,s);
  str(o2a:4,s);
  outtextxy1(392,93,s);
  str(r2b:4,s);
  outtextxy1(184,54,s);
  str(r2c:4,s);
  outtextxy1(280,54,s);
  str(r2a:4,s);
  outtextxy1(376,54,s);
  str(p1b:4,s);
  outtextxy1(141,244,s);
  str(e1b:4,s);
  outtextxy1(167,244,s);
  str(o1b:4,s);
  outtextxy1(195,244,s);
  str(p1c:4,s);
  outtextxy1(245,244,s);
  str(e1c:4,s);
  outtextxy1(272,244,s);
  str(o1c:4,s);
  outtextxy1(299,244,s);
  str(p1a:4,s);
  outtextxy1(343,244,s);
  str(e1a:4,s);
  outtextxy1(372,244,s);
  str(o1a:4,s);
  outtextxy1(398,244,s);
  str(r1b:4,s);
  outtextxy1(180,201,s);
  str(r1c:4,s);
  outtextxy1(284,202,s);
  str(r1a:4,s);
  outtextxy1(382,201,s);
  str(p1e:4,s);
  outtextxy1(18,136,s);
  str(e1e:4,s);
  outtextxy1(45,136,s);
  str(o1e:4,s);
  outtextxy1(72,136,s);
  str(p1d:4,s);
  outtextxy1(22,244,s);
  str(e1d:4,s);
  outtextxy1(49,244,s);
  str(o1d:4,s);
  outtextxy1(76,244,s);
  str(r1e:4,s);
  outtextxy1(61,94,s);
  str(r1d:4,s);
  outtextxy1(61,201,s);
 end;
end;

procedure saveedit;
begin
 assignfile(paramfile,'params');
 rewrite(paramfile);
 write(paramfile,g1a,g1b,g1c,g1d,g1e,s1a,s1b,s1c,s1d,s1e,g2a,g2b,g2c);
 closefile(paramfile);
end;

procedure initprogram;
begin
 clrscr;
 pr := 150; pt := 0;
 getfunctions;
 assignfile(paramfile,'params');
 {$i-}
 reset(paramfile);
 {$i+}
 if IOResult <> 0 then
  begin
   g1a :=  20;  s1a := 21;
   g1b :=  20;  s1b := 21;
   g1c :=  20;  s1c := 21;
   g1d :=   4;  s1d := 6;
   g1e :=   4;  s1e := 6;
   g2a :=   10;  g2b :=  10; g2c := 10;
   rewrite(paramfile);
   write(paramfile,g1a,g1b,g1c,g1d,g1e,s1a,s1b,s1c,s1d,s1e,g2a,g2b,g2c);
  end
 else read(paramfile,g1a,g1b,g1c,g1d,g1e,s1a,s1b,s1c,s1d,s1e,g2a,g2b,g2c);
 loadedit(g2a,g2b,g2c,g1a,s1a,g1b,s1b,g1c,s1c,g1d,s1d,g1e,s1e);

 closefile(paramfile);
 phi := 1536; theta := 0;
 gazex := 100; gazey := -100;
 phi2 := gazex;
 theta2 := gazey;
 initialize;
 setwritemode(0);
 toptrace := round(0.7*vsize);
 middletrace := round(0.8*vsize);
 bottomtrace := round(0.9* vsize);
 makecube;
 makebody;
 makehead;
 makeeyes;
 fillchar(target2,sizeof(target2),$ff);
 fillchar(arm2,sizeof(arm2),$ff);
 fillchar(head2,sizeof(head2),$ff);
 fillchar(eyes2,sizeof(eyes2),$ff);
 rotatelook(head3,head3r,6,theta2,phi2);
 rotatelook(eyes3,eyes3r,10,theta2,phi2);
 seteyes;
 initialize;
 environment;
 setwritemode(0);    // set to write first time
 setpencolor(1);

 computeenvironment;
 drawenvironment;
 computearmhead;
 drawarmhead;

 setwritemode(1);  // ready to start XOring,  begining with erase.

 tottime := 20; tempt := 0; usediagram := false;
 ch := ' '; purpose := false; pt := 0;
 loaddiagram;
 elapsedtime := 0.0;
 stage := 1;
end;

 { main loop }
procedure runprogram;
begin
  if purpose then
   begin
    pt := (pt + 30);
    if pt > 2047 then pt := -2048;
   end;

  if phi > maxmin1 then phi := phi - 2 * one80;
  if phi < -one80 then phi := phi + 2 * one80;
  if theta > maxmin1 then theta := theta - 2 * one80;
  if theta < -one80 then theta := theta + 2 * one80;
  filldiagram;
  {CALCULATE THE STATE OF ENVIRONMENT AND ALL CONTROL SYSTEMS}
  writetoclip('A');
  environment;
  writetoclip('B');
  distance2;
  vertical2;
  azimuth2;
  link2to1;
  writetoclip('C');

  distance1;
  vertical1;
  azimuth1;
  gaze1x;
  gaze1y;
  link1to0;
  writetoclip('D');

  {SHOW RUNNING PLOTS OF DATA}
  setwritemode(0);
  if (not usediagram) and (not editing) then
  begin
   if tottime = 20 then
    begin
     outtextxy(20,toptrace - 35,'VISUAL DEVIATION OF FINGER FROM TARGET');
     outtextxy(20,toptrace,'Perceived Deviation Y,P2B');
     outtextxy(20,middletrace,'Perceived Deviation X, P2A');
     outtextxy(20,bottomtrace,'Perceived Deviation Z, P2C');
    end;
   begin
    putpixel(tottime,toptrace - p2a div 7,1);
    putpixel(tottime,middletrace - p2b div 7,1);
    putpixel(tottime,bottomtrace - p2c div 20,1);
    putpixel(tottime,toptrace,1);
    putpixel(tottime,middletrace,1);
    putpixel(tottime,bottomtrace,1);
    inc(tottime);
   end;
   if (tottime >= hcenter) then
    begin
     setviewport(0,330,hcenter,vsize - 1);
     clearviewport; tottime := 20;
    end;
  end; // of running plots

  phi2 := gazex;          // set eye angles
  theta2 := gazey;

  { ERASE, RECOMPUTE, AND REDRAW LITTLE MAN'S HEAD,
    ARM, TARGET }

  if ch <> 'q' then
  begin
    setpencolor(0);

    drawenvironment;        // erase environment and arm
    drawarmhead;                              // erase movable objects
    writetoclip('D');

    rotatelook(head3,head3r,6,theta2,phi2);   // recompute eye angles
    rotatelook(eyes3,eyes3r,10,theta2,phi2);
    seteyes;
    writetoclip('E');

    computeenvironment;     // compute changes in viewpoint
    computearmhead;         // recompute arm, head, target
    setpencolor(1);
    drawenvironment;        // redraw environment and arm
    drawarmhead;            // draw moveable objects
    writetoclip('F');

  end;
end;

procedure iterate;
begin
 case stage of
  0: initprogram;
  1: runprogram;
  2: begin
      setviewport(0,330,hcenter,vsize - 1);
      clearviewport; tottime := 20;
      stage := 1;
     end;
 end;
 elapsedtime := elapsedtime + 0.02;
end;

begin
end.


